"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import itertools
import logging
import os
import time
from collections import deque
from typing import TYPE_CHECKING, Any, Deque, Dict, Iterable, Iterator, List, Optional, Sequence, Set, Tuple

from inmanta import plugins
from inmanta.ast import Anchor, AnchorTarget, CompilerException, CycleException, Location, MultiException, RuntimeException
from inmanta.ast.attribute import RelationAttribute
from inmanta.ast.entity import Entity, Implementation
from inmanta.ast.statements import DefinitionStatement, TypeDefinitionStatement
from inmanta.ast.statements.define import DefineEntity, DefineImplement, DefineIndex, DefineRelation, DefineTypeConstraint
from inmanta.ast.type import TYPES, Type
from inmanta.const import LOG_LEVEL_TRACE
from inmanta.execute.proxy import UnsetException
from inmanta.execute.runtime import (
    DelayedResultVariable,
    ExecutionContext,
    ExecutionUnit,
    Instance,
    QueueScheduler,
    RelationAttributeVariable,
    Resolver,
    ResultVariableProxy,
    VariableABC,
    Waiter,
)
from inmanta.execute.tracking import ModuleTracker

if TYPE_CHECKING:
    from inmanta.ast import BasicBlock, NamedType, Statement  # noqa: F401
    from inmanta.compiler import Compiler  # noqa: F401
    from inmanta.module import RelationPrecedenceRule


DEBUG = True
LOGGER = logging.getLogger(__name__)


MAX_ITERATIONS = 10000


class Scheduler(object):
    """
    This class schedules statements for execution
    """

    def __init__(
        self, track_dataflow: bool = False, relation_precedence_rules: Optional[List["RelationPrecedenceRule"]] = None
    ) -> None:
        if relation_precedence_rules is None:
            relation_precedence_rules = []
        self.track_dataflow: bool = track_dataflow
        self.types: Dict[str, Type] = {}
        # The precedence rules specified in the project.yml file. This list may contain rules that are invalid with
        # respect to the model.
        self.relation_precedence_rules: List["RelationPrecedenceRule"] = relation_precedence_rules

    def _set_precedence_rules_on_relationship_attributes(self) -> List[RelationAttribute]:
        """
        This method:
           * Validates the relation precedence rules in self.relation_precedence_policy and raises an exception for invalid
             relation precedence rules.
           * Set the RelationAttribute.relation_precedence_rules field on all RelationAttributes.
           * Returns the list of RelationAttributes for which a relation precedence policy exists saying that it should
             be frozen before another RelationshipAttribute.
        """
        if not self.relation_precedence_rules:
            return []
        LOGGER.warning(
            "[EXPERIMENTAL FEATURE] Using the relation precedence policy defined in the project.yml file to determine list "
            "freeze order."
        )
        result = []
        for rule in self.relation_precedence_rules:
            LOGGER.info("Loaded relation precedence rule: %s", rule)
            first_attribute: RelationAttribute = self._get_relation_attribute_with_precedence_rule(
                rule.first_type, rule.first_relation_name
            )
            then_attribute: RelationAttribute = self._get_relation_attribute_with_precedence_rule(
                rule.then_type, rule.then_relation_name
            )
            first_attribute.add_freeze_dependent(then_attribute)
            if first_attribute not in result:
                result.append(first_attribute)
        return result

    def _get_relation_attribute_with_precedence_rule(self, entity_type_name: str, relationship_name: str) -> RelationAttribute:
        """
        Return the RelationAttribute in `self.types` that has the given `entity_type_name` and `relationship_name`.
        """
        if not self.types:
            raise Exception("The self.define_types() method should be called first")
        if entity_type_name not in self.types:
            raise InvalidRelationPrecedenceRuleError(
                f"A relation precedence rule was defined for {entity_type_name}, but no such type was defined"
            )
        current_type: Type = self.types[entity_type_name]
        if not isinstance(current_type, Entity):
            raise InvalidRelationPrecedenceRuleError(
                f"A relation precedence rule was defined for non-entity type {current_type}"
            )
        assert isinstance(current_type, Entity)  # Make mypy happy
        attributes_of_entity = current_type.attributes
        if relationship_name not in attributes_of_entity:
            raise InvalidRelationPrecedenceRuleError(
                f"A relation precedence rule was defined for {entity_type_name}.{relationship_name}, "
                f"but entity {entity_type_name} doesn't have an attribute {relationship_name}.",
            )
        attribute = attributes_of_entity[relationship_name]
        if not isinstance(attribute, RelationAttribute):
            raise InvalidRelationPrecedenceRuleError(
                f"A relation precedence rule was defined for {entity_type_name}.{relationship_name}, "
                f"but attribute {relationship_name} is not a relationship attribute.",
            )
        return attribute

    def freeze_all(self, exns: List[CompilerException]) -> None:
        for t in [t for t in self.types.values() if isinstance(t, Entity)]:
            t.final(exns)

        instances: List[Instance] = self.types["std::Entity"].get_all_instances()

        for i in instances:
            i.final(exns)

    def dump(self, type: str = "std::Entity") -> None:
        instances: List[Instance] = self.types[type].get_all_instances()

        for i in instances:
            i.dump()

    def verify_done(self) -> List[Instance]:
        instances: List[Instance] = self.types["std::Entity"].get_all_instances()
        notdone = []
        for i in instances:
            if not i.verify_done():
                notdone.append(i)

        return notdone

    def get_types(self) -> Dict[str, Type]:
        return self.types

    def dump_not_done(self) -> None:
        for i in self.verify_done():
            i.dump()

    def sort_entities(self, entity_map: Dict[str, DefineEntity]) -> List[DefineEntity]:
        out: List[DefineEntity] = []
        loopstack: Set[str] = set()
        while len(entity_map) > 0:
            workon = next(iter(entity_map.keys()))
            self.do_sort_entities(entity_map, workon, out, loopstack)
        return out

    def do_sort_entities(
        self, entity_map: Dict[str, DefineEntity], name: str, acc: List[DefineEntity], loopstack: Set[str]
    ) -> None:
        nexte = entity_map[name]
        try:
            del entity_map[name]
            loopstack.add(name)
            for p in nexte.get_full_parent_names():
                if p in loopstack:
                    raise CycleException(nexte, p)
                if p in entity_map:
                    self.do_sort_entities(entity_map, p, acc, loopstack)
            loopstack.remove(name)
            acc.append(nexte)
        except CycleException as ce:
            ce.add(nexte)
            raise

    def define_types(self, compiler: "Compiler", statements: Sequence["Statement"], blocks: Sequence["BasicBlock"]) -> None:
        """
        This is the first compiler stage that defines all types_and_impl
        """
        # get all relevant stmts
        definitions: List["DefinitionStatement"] = [d for d in statements if isinstance(d, DefinitionStatement)]
        others: List["Statement"] = [d for d in statements if not isinstance(d, DefinitionStatement)]

        if not len(others) == 0:
            raise Exception("others not empty %s" % repr(others))

        # collect all  types and impls
        types_and_impl = {}

        # set primitive types
        compiler.get_ns().set_primitives(TYPES)

        # all stmts contributing types and impls
        newtypes: List[Tuple[str, NamedType]] = [
            k for k in [t.register_types() for t in definitions if isinstance(t, TypeDefinitionStatement)] if k is not None
        ]

        for name, type_symbol in newtypes:
            types_and_impl[name] = type_symbol

        # now that we have objects for all types, populate them
        implements = [t for t in definitions if isinstance(t, DefineImplement)]
        other_definitions = [t for t in definitions if not isinstance(t, DefineImplement)]
        entities: Dict[str, DefineEntity] = {t.fullName: t for t in other_definitions if isinstance(t, DefineEntity)}
        type_constraints = [t for t in other_definitions if isinstance(t, DefineTypeConstraint)]
        other_definitions = [t for t in other_definitions if not isinstance(t, (DefineEntity, DefineTypeConstraint))]
        indices = [t for t in other_definitions if isinstance(t, DefineIndex)]
        other_definitions = [t for t in other_definitions if not isinstance(t, DefineIndex)]

        # first type constraints so attribute defaults can be type checked
        for tc in type_constraints:
            tc.evaluate()

        # then entities, so we have inheritance
        # parents first
        for entity in self.sort_entities(entities):
            entity.evaluate()

        for other in other_definitions:
            other.evaluate()

        # indices late, as they require all attributes
        for index in indices:
            index.evaluate()

        # lastly the implements, as they require implementations
        for implement in implements:
            implement.evaluate()

        compiler.plugins = {k: v for k, v in types_and_impl.items() if isinstance(v, plugins.Plugin)}
        types = {k: v for k, v in types_and_impl.items() if isinstance(v, Type)}

        # normalize plugins
        for p in compiler.plugins.values():
            p.normalize()

        # give type info to all types, to normalize blocks inside them
        # normalize implementations last because they have subblocks that might depend on other type information
        for t in sorted(types.values(), key=lambda t: isinstance(t, Implementation)):
            t.normalize()

        # normalize root blocks
        for block in blocks:
            block.normalize()

        self.types = {k: v for k, v in types_and_impl.items() if isinstance(v, Type)}

    def get_anchormap(
        self, compiler: "Compiler", statements: Sequence["Statement"], blocks: Sequence["BasicBlock"]
    ) -> Sequence[Tuple[Location, AnchorTarget]]:
        prev = time.time()

        # first evaluate all definitions, this should be done in one iteration
        self.define_types(compiler, statements, blocks)

        # relations are also in blocks
        not_relation_statements: Iterator[Statement] = (s for s in statements if not isinstance(s, DefineRelation))

        anchors: Iterator[Anchor] = (
            anchor
            for container in itertools.chain(not_relation_statements, blocks)  # container: Union[Statement, BasicBlock]
            for anchor in container.get_anchors()
            if anchor is not None
        )

        range_to_anchor_target = [(anchor.get_location(), anchor.resolve()) for anchor in anchors]
        range_to_anchor_target = [(f, t) for f, t in range_to_anchor_target if t is not None]

        now = time.time()
        LOGGER.debug("Anchormap took %f seconds", now - prev)

        return range_to_anchor_target

    def anchormap(
        self, compiler: "Compiler", statements: Sequence["Statement"], blocks: Sequence["BasicBlock"]
    ) -> Sequence[Tuple[Location, Location]]:
        """
        This methode exists for backward compatibility with inmantals
        """
        range_to_anchor_target = self.get_anchormap(compiler, statements, blocks)
        range_to_range = [(f, t.location) for f, t in range_to_anchor_target]

        return range_to_range

    def find_wait_cycle(self, attributes_with_precedence_rule: List[RelationAttribute], allwaiters: Set[Waiter]) -> bool:
        """
        Preconditions: no progress is made anymore

        This means that all DelayedResultVariable that have not been frozen either have
        no progress potential or have outstanding promises.
        In this case, all gradual execution has been executed to the maximal extent.

        Any DelayedResultVariable that still has progress potential has at least one waiter that doesn't support gradual
        execution (e.g. a plugin).
        If it takes input from another DelayedResultVariables, it will still have this promise outstanding.

        The compiler will not freeze the DelayedResultVariable with progress potential because it has an outstanding promise.
        The compiler will not freeze the other because it has no progress potential.

        This causes the compiler to be stuck.

        The root cause is that progress potential is only calculated locally.

        For performance reasons, we keep progress potential local and instead detect this situation here.
        """

        def resolve_proxies(variable: Optional[VariableABC]) -> Optional[VariableABC]:
            if variable is None or not isinstance(variable, ResultVariableProxy):
                return variable
            return resolve_proxies(variable.variable)

        # Determine drvs that should be frozen to break the cycle
        freeze_candidates: List[DelayedResultVariable[object]] = []
        for waiter in allwaiters:
            for rv in waiter.requires.values():
                real_rv: Optional[VariableABC] = resolve_proxies(rv)
                if isinstance(real_rv, DelayedResultVariable):
                    if real_rv.hasValue:
                        # get_progress_potential fails when there is a value already
                        continue
                    if real_rv.get_waiting_providers() > 0 and real_rv.get_progress_potential() > 0:
                        freeze_candidates.append(real_rv)

        if not freeze_candidates:
            return False
        # Use the relation precedence rules to determine which drv should be frozen
        queue = PrioritisedDelayedResultVariableQueue(attributes_with_precedence_rule, freeze_candidates)
        drv_to_freeze = queue.popleft()
        LOGGER.debug("Waiting blocked on %s", drv_to_freeze)
        drv_to_freeze.freeze()
        return True

    def run(self, compiler: "Compiler", statements: Sequence["Statement"], blocks: Sequence["BasicBlock"]) -> bool:
        """
        Evaluate the current graph
        """
        prev = time.time()
        start = prev

        # first evaluate all definitions, this should be done in one iteration
        self.define_types(compiler, statements, blocks)
        attributes_with_precedence_rule: List[RelationAttribute] = self._set_precedence_rules_on_relationship_attributes()

        # give all loose blocks an empty XC
        # register the XC's as scopes
        # All named scopes are now present

        for block in blocks:
            res = Resolver(block.namespace, self.track_dataflow)
            xc = ExecutionContext(block, res)
            block.context = xc
            block.namespace.scope = xc
            block.warn_shadowed_variables()

        # setup queues
        # queue for runnable items
        basequeue: Deque[Waiter] = deque()
        # queue for RV's that are delayed
        waitqueue = PrioritisedDelayedResultVariableQueue(attributes_with_precedence_rule)
        # queue for RV's that are delayed and had no effective waiters when they were first in the waitqueue
        zerowaiters: Deque[DelayedResultVariable[Any]] = deque()
        # queue containing everything, to find hanging statements
        all_statements: Set[Waiter] = set()

        # Wrap in object to pass around
        queue = QueueScheduler(compiler, basequeue, waitqueue, self.types, all_statements)

        # emit all top level statements
        for block in blocks:
            block.context.emit(queue.for_tracker(ModuleTracker(block)))

        # start an evaluation loop
        i = 0
        count = 0
        max_iterations = int(os.getenv("INMANTA_MAX_ITERATIONS", MAX_ITERATIONS))
        while i < max_iterations:
            now = time.time()

            # check if we can stop the execution
            if len(basequeue) == 0 and len(waitqueue) == 0 and len(zerowaiters) == 0:
                break
            else:
                i += 1

            LOGGER.debug(
                "Iteration %d (e: %d, w: %d, p: %d, done: %d, time: %f)",
                i,
                len(basequeue),
                len(waitqueue),
                len(zerowaiters),
                count,
                now - prev,
            )
            prev = now

            # evaluate all that is ready
            while len(basequeue) > 0:
                next = basequeue.popleft()
                try:
                    next.execute()
                    all_statements.discard(next)
                    count = count + 1
                except UnsetException as e:
                    # some statements don't know all their dependencies up front,...
                    next.requeue_with_additional_requires(object(), e.get_result_variable())

            # all safe stmts are done
            progress = False
            assert not basequeue

            # find a RV that has waiters, so freezing creates progress
            while len(waitqueue) > 0 and not progress:
                next_rv = waitqueue.popleft()
                if next_rv.hasValue:
                    # already froze itself
                    continue
                if next_rv.get_progress_potential() <= 0:
                    zerowaiters.append(next_rv)
                elif next_rv.get_waiting_providers() > 0:
                    # definitely not done
                    # drop from queue
                    # will requeue when value is added
                    next_rv.unqueue()
                else:
                    # freeze it and go to next iteration, new statements will be on the basequeue
                    LOGGER.log(LOG_LEVEL_TRACE, "Freezing %s", next_rv)
                    next_rv.freeze()
                    progress = True

            # no waiters in waitqueue,...
            # see if any zerowaiters have become gotten waiters
            if not progress:
                zerowaiters_tmp = [w for w in zerowaiters if not w.hasValue]
                waitqueue.replace(w for w in zerowaiters_tmp if w.get_progress_potential() > 0)
                zerowaiters = deque(w for w in zerowaiters_tmp if w.get_progress_potential() <= 0)
                while len(waitqueue) > 0 and not progress:
                    LOGGER.debug("Moved zerowaiters to waiters")
                    next_rv = waitqueue.popleft()
                    if next_rv.get_waiting_providers() > 0:
                        next_rv.unqueue()
                    else:
                        LOGGER.log(LOG_LEVEL_TRACE, "Freezing %s", next_rv)
                        next_rv.freeze()
                        progress = True

            if not progress:
                # nothing works anymore, attempt to unfreeze wait cycle
                progress = self.find_wait_cycle(attributes_with_precedence_rule, queue.allwaiters)

            if not progress:
                # no one waiting anymore, all done, freeze and finish
                LOGGER.debug("Finishing statements with no waiters")

                while len(zerowaiters) > 0:
                    next_rv = zerowaiters.pop()
                    next_rv.freeze()

        now = time.time()
        LOGGER.debug(
            "Iteration %d (e: %d, w: %d, p: %d, done: %d, time: %f)",
            i,
            len(basequeue),
            len(waitqueue),
            len(zerowaiters),
            count,
            now - prev,
        )

        if i == max_iterations:
            raise CompilerException(f"Could not complete model, max_iterations {max_iterations} reached.")

        excns: List[CompilerException] = []
        self.freeze_all(excns)

        now = time.time()
        LOGGER.info(
            "Total compilation time %f",
            now - start,
        )

        if len(excns) == 0:
            pass
        elif len(excns) == 1:
            raise excns[0]
        else:
            raise MultiException(excns)

        if all_statements:
            stmt = None
            for st in all_statements:
                if isinstance(st, ExecutionUnit):
                    stmt = st
                    break

            assert stmt is not None

            raise RuntimeException(stmt.expression, "not all statements executed %s" % all_statements)

        return True


class InvalidRelationPrecedenceRuleError(CompilerException):
    """
    A CompilerException that is raised when the project.yml file
    contains a relation precedence rule that invalid with respect
    to the given project.
    """

    def __init__(self, msg: str) -> None:
        super(InvalidRelationPrecedenceRuleError, self).__init__(msg)


class PrioritisedDelayedResultVariableQueue:
    """
    A queue for DelayedResultVariables that is prioritized based on the
    relation precedence policy passed to the Compiler. This queue will return elements
    in the following order:

    * First return the DelayedResultVariables for relations that do not have an order constraint.
    * Then return DelayedResultVariables for relations with order constraint.
      They are returned in an order that is valid with respect to the constraints.
    * Finally, all DelayedResultVariables that are not associated with an entity are
      returned.
    """

    def __init__(
        self,
        attributes_with_precedence_rule: List[RelationAttribute],
        drvs: Optional[List[DelayedResultVariable[object]]] = None,
    ) -> None:
        relation_precedence_graph = RelationPrecedenceGraph(attributes_with_precedence_rule)
        # A queue that indicates a valid order in which the self._constraint_variables have to be returned
        # This queue is never modified.
        self._freeze_order: Deque[RelationAttribute] = deque(relation_precedence_graph.get_freeze_order())
        # Copy of self._freeze_order. At all times the first element of this queue
        # points to the next type that should be returned from self._constraint_variables
        self._freeze_order_working_list: Deque[RelationAttribute] = self._freeze_order.copy()

        self._unconstraint_variables: Deque[DelayedResultVariable[object]] = deque()
        self._constraint_variables: Dict[RelationAttribute, Deque[DelayedResultVariable[object]]] = {
            relation_attribute: deque() for relation_attribute in self._freeze_order
        }
        self._non_relation_variables: Deque[DelayedResultVariable] = deque()

        # Populate queue with given DelayedResultVariables
        drvs = drvs if drvs else []
        for drv in drvs:
            self.append(drv, dont_reset_working_list=True)

    def __len__(self) -> int:
        """
        Return the number of elements present in this queue.
        """
        return (
            len(self._unconstraint_variables)
            + sum(len(v) for v in self._constraint_variables.values())
            + len(self._non_relation_variables)
        )

    def append(self, drv: DelayedResultVariable[object], dont_reset_working_list: bool = False) -> None:
        """
        Append on the right side of the queue.

        :param dont_reset_working_list: This argument exists to increase performance by preventing
                                        unnecessary resets of the `self._freeze_order_working_list` queue.
        """
        if not isinstance(drv, RelationAttributeVariable):
            self._non_relation_variables.append(drv)
        elif drv.attribute in self._constraint_variables:
            self._constraint_variables[drv.attribute].append(drv)
            if not dont_reset_working_list and drv.attribute not in self._freeze_order_working_list:
                # working list is dirty
                self._freeze_order_working_list = self._freeze_order.copy()
        else:
            self._unconstraint_variables.append(drv)

    def popleft(self) -> DelayedResultVariable[Any]:
        """
        Remove element from the left side of the queue and return it.
        """
        try:
            return self._unconstraint_variables.popleft()
        except IndexError:
            # Empty
            pass
        try:
            return self._get_next_constraint_variable()
        except IndexError:
            # Empty
            pass
        return self._non_relation_variables.popleft()

    def _get_next_constraint_variable(self) -> DelayedResultVariable[object]:
        if not self._constraint_variables:
            raise IndexError()
        while self._freeze_order_working_list:
            relation_attribute: RelationAttribute = self._freeze_order_working_list[0]
            if relation_attribute not in self._constraint_variables or not self._constraint_variables[relation_attribute]:
                self._freeze_order_working_list.popleft()
            else:
                return self._constraint_variables[relation_attribute].popleft()
        raise IndexError()

    def replace(self, drvs: Iterable[DelayedResultVariable[object]]) -> None:
        """
        Remove all elements from this queue and add the elements provided in drvs.
        """
        self._unconstraint_variables.clear()
        for queue in self._constraint_variables.values():
            queue.clear()
        self._non_relation_variables.clear()
        for drv in drvs:
            self.append(drv, dont_reset_working_list=True)
        self._freeze_order_working_list = self._freeze_order.copy()


class CycleInRelationPrecedencePolicyError(CompilerException):
    """
    Raised when a cycle exists in the relation precedence rules provided to the compiler.
    """

    def __init__(self) -> None:
        super(CycleInRelationPrecedencePolicyError, self).__init__("A cycle exists in the relation precedence policy")


class RelationPrecedenceGraph:
    """
    A graph representation of the relation precedence policy provided to the compiler.
    """

    def __init__(self, relation_attributes_with_precedence_rule: Optional[List[RelationAttribute]] = None) -> None:
        if relation_attributes_with_precedence_rule is None:
            relation_attributes_with_precedence_rule = []
        # The root nodes of the graph, where all other nodes attach to.
        self.root_nodes: Set[RelationPrecedenceGraphNode] = set()
        self.attribute_to_node: Dict[RelationAttribute, RelationPrecedenceGraphNode] = {}
        # Creates nodes in graph
        for first_attribute in relation_attributes_with_precedence_rule:
            for then_attribute in first_attribute.freeze_dependents:
                self.add_precedence_rule(first_attribute, then_attribute)

    def add_precedence_rule(self, first_attribute: RelationAttribute, then_attribute: RelationAttribute) -> None:
        """
        Add a rule that `first_attribute` should be frozen before `then_attribute`.
        """
        first_node = self._get_or_create_node(first_attribute, attach_to_root=True)
        then_node = self._get_or_create_node(then_attribute, attach_to_root=False)
        if then_node in self.root_nodes:
            self.root_nodes.remove(then_node)
        first_node.add_dependent(then_node)

    def _get_or_create_node(self, relation_attribute: RelationAttribute, attach_to_root: bool) -> "RelationPrecedenceGraphNode":
        """
        Get the RelationPrecedenceGraphNode for the given relation_attribute in this graph
        or create a new node if no such node exists.

        :param attach_to_root: true iff attach the node to the root nodes if the node was newly created.
        """
        if relation_attribute not in self.attribute_to_node:
            node = RelationPrecedenceGraphNode(relation_attribute)
            self.attribute_to_node[relation_attribute] = node
            if attach_to_root:
                self.root_nodes.add(node)
        else:
            node = self.attribute_to_node[relation_attribute]
        return node

    def get_freeze_order(self) -> List[RelationAttribute]:
        """
        Return all the RelationAttributes in this graph in the order in which
        they should be frozen.
        """
        if not self.attribute_to_node:
            return []
        work: Set[RelationPrecedenceGraphNode] = set(self.root_nodes)
        result: List[RelationAttribute] = []

        def get_next_ready_item_in_work() -> RelationPrecedenceGraphNode:
            assert work
            for current_node in work:
                if all(dep.relation_attribute in result for dep in current_node.dependencies):
                    return current_node
            raise CycleInRelationPrecedencePolicyError()

        while work:
            node: RelationPrecedenceGraphNode = get_next_ready_item_in_work()
            work.remove(node)
            if node.relation_attribute in result:
                raise CycleInRelationPrecedencePolicyError()
            result.append(node.relation_attribute)
            work.update(node.dependents)

        if len(result) != len(self.attribute_to_node):
            raise CycleInRelationPrecedencePolicyError()
        return result


class RelationPrecedenceGraphNode:
    """
    A node in the RelationPrecedenceGraph that represents the relationship of an Inmanta entity.
    """

    def __init__(self, relation_attribute: RelationAttribute) -> None:
        self.relation_attribute: RelationAttribute = relation_attribute
        self.dependents: Set[RelationPrecedenceGraphNode] = set()
        self.dependencies: Set[RelationPrecedenceGraphNode] = set()

    def add_dependent(self, dependent: "RelationPrecedenceGraphNode") -> None:
        self.dependents.add(dependent)
        dependent.dependencies.add(self)



"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import typing
from collections import abc
from dataclasses import dataclass
from typing import Optional

from asyncpg import Connection


async def update(connection: Connection) -> None:
    """
    Recreate resource state types without the processing_events state.
    """
    resource_state_values: abc.Sequence[str] = [
        "unavailable",
        "skipped",
        "dry",
        "deployed",
        "failed",
        "deploying",
        "available",
        "cancelled",
        "undefined",
        "skipped_for_undefined",
    ]
    await replace_enum_type(
        EnumDefinition(
            name="resourcestate",
            values=resource_state_values,
            deleted_values={"processing_events": "deploying"},
            columns={
                "public.resource": [ColumnDefinition(name="status", default="available")],
                "public.resourceaction": [ColumnDefinition(name="status", default="available")],
            },
        ),
        connection=connection,
    )
    await replace_enum_type(
        EnumDefinition(
            name="non_deploying_resource_state",
            values=[v for v in resource_state_values if v != "deploying"],
            # migrate values to be safe even though processing_events state was actually unreachable in practice for this type
            deleted_values={"processing_events": "available"},
            columns={"public.resource": [ColumnDefinition(name="last_non_deploying_status", default="available")]},
        ),
        connection=connection,
    )


class ColumnDefinition(typing.NamedTuple):
    name: str
    default: Optional[str]


@dataclass(frozen=True)
class EnumDefinition:
    name: str
    values: abc.Sequence[str]
    deleted_values: abc.Mapping[str, Optional[str]]  # deleted values mapped to new value if any currently exist
    columns: abc.Mapping[str, abc.Sequence[ColumnDefinition]]  # columns with defaults


async def replace_enum_type(new_type: EnumDefinition, *, connection: Connection) -> None:
    """
    Completely replaces an enum type with a new definition with the same name.

    :param new_type: The definition of the new type. Assumed to be an internal construct, this method is not safe against
        injections via this object's attributes.
    """
    temp_name: str = f"_old_{new_type.name}"
    await connection.execute(
        f"""
        ALTER TYPE {new_type.name} RENAME TO {temp_name};
        CREATE TYPE {new_type.name} AS ENUM(%s);
        """
        % (", ".join(f"'{v}'" for v in new_type.values))
    )
    for table, columns in new_type.columns.items():
        for column, default in columns:
            for old_value, new_value in new_type.deleted_values.items():
                await connection.execute(f"UPDATE {table} SET {column}=$1 WHERE {column}=$2", new_value, old_value)
            await connection.execute(f"ALTER TABLE {table} ALTER COLUMN {column} DROP DEFAULT")
            await connection.execute(
                # can't cast directly between enums -> go via varchar
                f"ALTER TABLE {table} ALTER COLUMN {column} TYPE {new_type.name} USING {column}::varchar::{new_type.name}"
            )
            await connection.execute(f"ALTER TABLE {table} ALTER COLUMN {column} SET DEFAULT '{default}'")
    await connection.execute(f"DROP TYPE {temp_name}")



"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from asyncpg import Connection

DISABLED = False


async def update(connection: Connection) -> None:
    schema = """
    CREATE TYPE notificationseverity AS ENUM('message', 'info', 'success', 'warning', 'error');

    -- Table: public.notification
    CREATE TABLE IF NOT EXISTS public.notification (
        id uuid NOT NULL,
        environment uuid NOT NULL REFERENCES environment(id) ON DELETE CASCADE,
        created TIMESTAMP WITH TIME ZONE NOT NULL,
        title varchar NOT NULL,
        message varchar NOT NULL,
        severity notificationseverity DEFAULT 'message',
        uri varchar NOT NULL,
        read boolean NOT NULL DEFAULT FALSE,
        cleared boolean NOT NULL DEFAULT FALSE,
        PRIMARY KEY(environment, id)
    );
    CREATE INDEX IF NOT EXISTS notification_env_created_id_index ON notification (environment, created DESC, id);
    """
    async with connection.transaction():
        await connection.execute(schema)



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
import re
import signal
import subprocess
import sys
from subprocess import TimeoutExpired
from threading import Timer

import pytest

import inmanta.util
from inmanta import const


def get_command(
    tmp_dir,
    stdout_log_level=None,
    log_file=None,
    log_level_log_file=None,
    timed=False,
    dbport=None,
    dbname="inmanta",
    dbhost=None,
    dbuser=None,
    dbpass=None,
    config_dir=None,
    server_extensions=[],
    version=False,
):
    root_dir = tmp_dir.mkdir("root").strpath
    log_dir = os.path.join(root_dir, "log")
    state_dir = os.path.join(root_dir, "data")
    for directory in [log_dir, state_dir]:
        os.mkdir(directory)
    config_file = os.path.join(root_dir, "inmanta.cfg")

    if dbport is not None:
        port = dbport
    else:
        port = inmanta.util.get_free_tcp_port()

    with open(config_file, "w+", encoding="utf-8") as f:
        f.write("[config]\n")
        f.write("log-dir=" + log_dir + "\n")
        f.write("state-dir=" + state_dir + "\n")
        f.write("[database]\n")
        f.write("port=" + str(port) + "\n")
        f.write("name=" + dbname + "\n")
        if dbhost:
            f.write(f"host={dbhost}\n")
        if dbuser:
            f.write(f"username={dbuser}\n")
        if dbpass:
            f.write(f"password={dbpass}\n")
        f.write("[server]\n")
        f.write(f"enabled_extensions={', '.join(server_extensions)}\n")

    args = [sys.executable, "-m", "inmanta.app"]
    if stdout_log_level:
        args.append("-" + "v" * stdout_log_level)
    if log_file:
        log_file = os.path.join(log_dir, log_file)
        args += ["--log-file", log_file]
    if log_file and log_level_log_file:
        args += ["--log-file-level", str(log_level_log_file)]
    if timed:
        args += ["--timed-logs"]
    if config_dir:
        args += ["--config-dir", config_dir]
    if version:
        args += ["--version"]
    args += ["-c", config_file, "server"]
    return (args, log_dir)


def do_run(args, env={}, cwd=None):
    baseenv = os.environ.copy()
    baseenv.update(env)
    process = subprocess.Popen(args, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=baseenv)
    return process


def convert_to_ascii(text):
    return [line for line in text.decode("ascii").split("\n") if line != ""]


def do_kill(process, killtime=3, termtime=2):
    def do_and_log(func, msg):
        def w():
            print(msg)
            func()

        return w

    t1 = Timer(killtime, do_and_log(process.kill, "killed process"))
    t2 = Timer(termtime, do_and_log(process.terminate, "terminated process"))
    t1.start()
    t2.start()

    out, err = process.communicate()

    t1.cancel()
    t2.cancel()

    stdout = convert_to_ascii(out)
    stderr = convert_to_ascii(err)
    return (stdout, stderr, process.returncode)


def run_without_tty(args, env={}, killtime=3, termtime=2):
    process = do_run(args, env)
    return do_kill(process, killtime, termtime)


def run_with_tty(args, killtime=3, termtime=2):
    """Could not get code for actual tty to run stable in docker, so we are faking it"""
    env = {const.ENVIRON_FORCE_TTY: "true"}
    return run_without_tty(args, env=env, killtime=killtime, termtime=termtime)


def get_timestamp_regex():
    return r"[\d]{4}\-[\d]{2}\-[\d]{2} [\d]{2}\:[\d]{2}\:[\d]{2}\,[\d]{3}"


def get_compiled_regexes(regexes, timed):
    result = []
    for regex in regexes:
        if timed:
            regex = get_timestamp_regex() + " " + regex
        compiled_regex = re.compile(regex)
        result.append(compiled_regex)
    return result


def is_colorama_package_available():
    try:
        import colorama  # noqa: F401
    except ModuleNotFoundError:
        return False
    return True


def test_verify_that_colorama_package_is_not_present():
    """
    The colorama package turns the colored characters in TTY-based terminal into uncolored characters.
    As such, this package should not be present.
    """
    assert not is_colorama_package_available()


@pytest.mark.parametrize_any(
    "log_level, timed, with_tty, regexes_required_lines, regexes_forbidden_lines",
    [
        (
            3,
            False,
            False,
            [r"[a-z.]*[ ]*INFO[\s]+Starting server endpoint", r"[a-z.]*[ ]*DEBUG[\s]+Starting Server Rest Endpoint"],
            [],
        ),
        (
            2,
            False,
            False,
            [r"[a-z.]*[ ]*INFO[\s]+Starting server endpoint"],
            [r"[a-z.]*[ ]*DEBUG[\s]+Starting Server Rest Endpoint"],
        ),
        (
            3,
            False,
            True,
            [
                r"\x1b\[32m[a-z.]*[ ]*INFO[\s]*\x1b\[0m\x1b\[34mStarting server endpoint",
                r"\x1b\[36m[a-z.]*[ ]*DEBUG[\s]*\x1b\[0m\x1b\[34mStarting Server Rest Endpoint",
            ],
            [],
        ),
        (
            2,
            False,
            True,
            [r"\x1b\[32m[a-z.]*[ ]*INFO[\s]*\x1b\[0m\x1b\[34mStarting server endpoint"],
            [r"\x1b\[36m[a-z.]*[ ]*DEBUG[\s]*\x1b\[0m\x1b\[34mStarting Server Rest Endpoint"],
        ),
        (
            3,
            True,
            False,
            [r"[a-z.]*[ ]*INFO[\s]+Starting server endpoint", r"[a-z.]*[ ]*DEBUG[\s]+Starting Server Rest Endpoint"],
            [],
        ),
        (
            2,
            True,
            False,
            [r"[a-z.]*[ ]*INFO[\s]+Starting server endpoint"],
            [r"[a-z.]*[ ]*DEBUG[\s]+Starting Server Rest Endpoint"],
        ),
        (
            3,
            True,
            True,
            [
                r"\x1b\[32m[a-z.]*[ ]*INFO[\s]*\x1b\[0m\x1b\[34mStarting server endpoint",
                r"\x1b\[36m[a-z.]*[ ]*DEBUG[\s]*\x1b\[0m\x1b\[34mStarting Server Rest Endpoint",
            ],
            [],
        ),
        (
            2,
            True,
            True,
            [r"\x1b\[32m[a-z.]*[ ]*INFO[\s]*\x1b\[0m\x1b\[34mStarting server endpoint"],
            [r"\x1b\[36m[a-z.]*[ ]*DEBUG[\s]*\x1b\[0m\x1b\[34mStarting Server Rest Endpoint"],
        ),
    ],
)
@pytest.mark.timeout(20)
def test_no_log_file_set(tmpdir, log_level, timed, with_tty, regexes_required_lines, regexes_forbidden_lines):
    if is_colorama_package_available() and with_tty:
        pytest.skip("Colorama is present")

    (args, log_dir) = get_command(tmpdir, stdout_log_level=log_level, timed=timed)
    if with_tty:
        (stdout, _, _) = run_with_tty(args)
    else:
        (stdout, _, _) = run_without_tty(args)
    log_file = "server.log"
    assert log_file not in os.listdir(log_dir)
    assert len(stdout) != 0
    check_logs(stdout, regexes_required_lines, regexes_forbidden_lines, timed)


@pytest.mark.parametrize_any(
    "log_level,with_tty, regexes_required_lines, regexes_forbidden_lines",
    [
        (
            3,
            False,
            [
                r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint",
                r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint",
            ],
            [],
        ),
        (
            "DEBUG",
            False,
            [
                r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint",
                r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint",
            ],
            [],
        ),
        (
            2,
            False,
            [r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint"],
            [r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint"],
        ),
        (
            "INFO",
            False,
            [r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint"],
            [r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint"],
        ),
        (
            3,
            True,
            [
                r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint",
                r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint",
            ],
            [],
        ),
        (
            2,
            True,
            [r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint"],
            [r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint"],
        ),
    ],
)
@pytest.mark.timeout(60)
def test_log_file_set(tmpdir, log_level, with_tty, regexes_required_lines, regexes_forbidden_lines):
    if is_colorama_package_available() and with_tty:
        pytest.skip("Colorama is present")

    log_file = "server.log"
    (args, log_dir) = get_command(tmpdir, log_file=log_file, log_level_log_file=log_level)
    if with_tty:
        (stdout, _, _) = run_with_tty(args)
    else:
        (stdout, _, _) = run_without_tty(args)
    assert log_file in os.listdir(log_dir)
    log_file = os.path.join(log_dir, log_file)
    with open(log_file, "r") as f:
        log_lines = f.readlines()
    check_logs(log_lines, regexes_required_lines, regexes_forbidden_lines, timed=True)
    check_logs(stdout, [], regexes_required_lines, timed=True)
    check_logs(stdout, [], regexes_required_lines, timed=False)


@pytest.mark.parametrize_any(
    "log_level, regexes_required_lines, regexes_forbidden_lines",
    [
        (
            5,
            [
                r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint",
                r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint",
            ],
            [],
        ),
        (
            4,
            [
                r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint",
                r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint",
            ],
            [],
        ),
        (
            3,
            [
                r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint",
                r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint",
            ],
            [],
        ),
        (
            2,
            [r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint"],
            [r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint"],
        ),
        (
            1,
            [],
            [
                r"[a-z.]*[ ]*DEBUG[\s]+[a-x\.A-Z]*[\s]Starting Server Rest Endpoint",
                r"[a-z.]*[ ]*INFO[\s]+[a-x\.A-Z]*[\s]Starting server endpoint",
            ],
        ),
    ],
)
@pytest.mark.timeout(60)
def test_log_stdout_log_level(tmpdir, log_level, regexes_required_lines, regexes_forbidden_lines):
    (args, log_dir) = get_command(tmpdir, stdout_log_level=log_level)
    (stdout, _, _) = run_without_tty(args)
    check_logs(stdout, regexes_required_lines, regexes_forbidden_lines, timed=False)


def check_logs(log_lines, regexes_required_lines, regexes_forbidden_lines, timed):
    compiled_regexes_requires_lines = get_compiled_regexes(regexes_required_lines, timed)
    compiled_regexes_forbidden_lines = get_compiled_regexes(regexes_forbidden_lines, timed)
    for line in log_lines:
        print(line)
    for regex in compiled_regexes_requires_lines:
        if not any(regex.match(line) for line in log_lines):
            pytest.fail("Required pattern was not found in log lines: %s" % (regex.pattern,))
    for regex in compiled_regexes_forbidden_lines:
        if any(regex.match(line) for line in log_lines):
            pytest.fail("Forbidden pattern found in log lines: %s" % (regex.pattern,))


def test_check_shutdown():
    process = do_run([sys.executable, os.path.join(os.path.dirname(__file__), "miniapp.py")])
    # wait for handler to be in place
    try:
        process.communicate(timeout=2)
    except TimeoutExpired:
        pass
    process.send_signal(signal.SIGUSR1)
    out, err, code = do_kill(process, killtime=3, termtime=1)
    print(out, err)
    assert code == 0
    assert "----- Thread Dump ----" in out
    assert "STOP" in out
    assert "SHUTDOWN COMPLETE" in out


def test_check_bad_shutdown():
    print([sys.executable, os.path.join(os.path.dirname(__file__), "miniapp.py"), "bad"])
    process = do_run([sys.executable, os.path.join(os.path.dirname(__file__), "miniapp.py"), "bad"])
    out, err, code = do_kill(process, killtime=5, termtime=2)
    print(out, err)
    assert code == 3
    assert "----- Thread Dump ----" in out
    assert "STOP" not in out
    assert "SHUTDOWN COMPLETE" not in out


def test_startup_failure(tmpdir, postgres_db, database_name):
    (args, log_dir) = get_command(
        tmpdir,
        dbport=postgres_db.port,
        dbname=database_name,
        dbhost=postgres_db.host,
        dbuser=postgres_db.user,
        dbpass=postgres_db.password,
        server_extensions=["badplugin"],
    )
    pp = ":".join(sys.path)
    # Add a bad module
    extrapath = os.path.join(os.path.dirname(__file__), "data", "bad_module_path")
    (stdout, stderr, code) = run_without_tty(args, env={"PYTHONPATH": pp + ":" + extrapath}, killtime=15, termtime=10)
    assert "inmanta                  ERROR   Server setup failed" in stdout
    assert (
        "                                 " + "inmanta.server.protocol.SliceStartupException: Slice badplugin.badslice "
        "failed to start because: Too bad, this plugin is broken"
    ) in stdout
    assert code == 4


@pytest.mark.parametrize("cache_cf_files", [True, False])
def test_compiler_exception_output(snippetcompiler, cache_cf_files):
    """
    This test case is also used to test the caching (issue 3838)
    Since this is a basic smoke test for argument parsing, no assertion
    about the caching is done here.
    """
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    number attr
end

implement Test using std::none

o = Test(attr="1234")
        """
    )
    cwd = snippetcompiler.project_dir if cache_cf_files else "."

    output = (
        f"""Could not set attribute `attr` on instance `__config__::Test (instantiated at {cwd}/main.cf:8)` """
        f"""(reported in Construct(Test) ({cwd}/main.cf:8))
caused by:
  Invalid value '1234', expected Number (reported in Construct(Test) ({cwd}/main.cf:8))
"""
    )

    def exec(*cmd):
        process = do_run([sys.executable, "-m", "inmanta.app"] + list(cmd), cwd=snippetcompiler.project_dir)
        _, err = process.communicate(timeout=30)
        assert output in err.decode()

    no_cache_option = [] if cache_cf_files else ["--no-cache"]

    cl_compile = ["compile"] + no_cache_option
    cl_export = ["export", "-J", "out.json"] + no_cache_option

    exec(*cl_compile)
    exec(*cl_export)


@pytest.mark.timeout(15)
@pytest.mark.parametrize_any(
    "cmd", [(["-X", "compile"]), (["compile", "-X"]), (["compile"]), (["export", "-X"]), (["-X", "export"]), (["export"])]
)
def test_minus_x_option(snippetcompiler, cmd):
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    nuber attr
end
"""
    )

    process = do_run([sys.executable, "-m", "inmanta.app"] + cmd, cwd=snippetcompiler.project_dir)
    out, err = process.communicate(timeout=30)
    if "-X" in cmd:
        assert "inmanta.ast.TypeNotFoundException: could not find type nuber in namespace" in str(err)
    else:
        assert "inmanta.ast.TypeNotFoundException: could not find type nuber in namespace" not in str(err)


@pytest.mark.timeout(20)
def test_warning_config_dir_option_on_server_command(tmpdir):
    non_existing_dir = os.path.join(tmpdir, "non_existing_dir")
    assert not os.path.isdir(non_existing_dir)
    (args, _) = get_command(tmpdir, stdout_log_level=3, config_dir=non_existing_dir)
    (stdout, _, _) = run_without_tty(args)
    stdout = "".join(stdout)
    assert "Starting server endpoint" in stdout
    assert f"Config directory {non_existing_dir} doesn't exist" in stdout


@pytest.mark.timeout(20)
def test_warning_min_c_option_file_doesnt_exist(snippetcompiler, tmpdir):
    non_existing_config_file = os.path.join(tmpdir, "non_existing_config_file")
    snippetcompiler.setup_for_snippet(
        """
entity Test:
    number attr
end
"""
    )
    config_options = ["-c", non_existing_config_file, "-vvv"]
    args = [sys.executable, "-m", "inmanta.app"] + config_options + ["compile"]
    process = do_run(args, cwd=snippetcompiler.project_dir)
    out, err = process.communicate(timeout=30)
    assert process.returncode == 0

    out = out.decode()
    err = err.decode()
    all_output = out + err

    assert "Starting compile" in all_output
    assert "Compile done" in all_output
    assert f"Config file {non_existing_config_file} doesn't exist" in all_output


@pytest.mark.parametrize_any(
    "with_tty, version_should_be_shown, regexes_required_lines, regexes_forbidden_lines",
    [
        (False, True, [r"Inmanta Service Orchestrator", r"Compiler version: ", r"Extensions:", r"\s*\* core:"], []),
        (True, True, [r"Inmanta Service Orchestrator", r"Compiler version: ", r"Extensions:", r"\s*\* core:"], []),
        (False, False, [], [r"Inmanta Service Orchestrator", r"Compiler version: ", r"Extensions:", r"\s*\* core:"]),
        (True, False, [], [r"Inmanta Service Orchestrator", r"Compiler version: ", r"Extensions:", r"\s*\* core:"]),
    ],
)
@pytest.mark.timeout(20)
def test_version_argument_is_set(tmpdir, with_tty, version_should_be_shown, regexes_required_lines, regexes_forbidden_lines):
    (args, log_dir) = get_command(tmpdir, version=version_should_be_shown)
    if with_tty:
        (stdout, _, _) = run_with_tty(args, killtime=15, termtime=10)
    else:
        (stdout, _, _) = run_without_tty(args, killtime=15, termtime=10)
    assert len(stdout) != 0
    check_logs(stdout, regexes_required_lines, regexes_forbidden_lines, False)


def test_init_project(tmpdir):
    args = [sys.executable, "-m", "inmanta.app", "project", "init", "-n", "test-project", "-o", tmpdir, "--default"]
    (stdout, stderr, return_code) = run_without_tty(args, killtime=15, termtime=10)
    test_project_path = os.path.join(tmpdir, "test-project")
    assert return_code == 0
    assert os.path.exists(test_project_path)
    (stdout, stderr, return_code) = run_without_tty(args, killtime=15, termtime=10)
    assert return_code != 0
    assert any("already exists" in error for error in stderr)



"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

"""
Tests to verify correctness/compatibility of code snippets in the docs.
"""

import os
import textwrap
import uuid
from collections import abc, defaultdict

import py
import pytest

from inmanta import data
from inmanta.data import model
from utils import v1_module_from_template

DOCS_DIR: str = os.path.join(os.path.dirname(__file__), "..", "docs")


@pytest.mark.slowtest
async def test_docs_snippet_partial_compile(
    tmpdir: py.path.local, snippetcompiler, modules_dir: str, server, client, environment: str
) -> None:
    """
    Verify that the partial compile example model is valid and is in fact equivalent to the full model it is compared to.
    """
    env_id: uuid.UUID = uuid.UUID(environment)
    snippets_dir: str = os.path.join(DOCS_DIR, "model_developers", "resource_sets")
    version: int

    handler_module_name: str = "host_handlers"
    v1_module_from_template(
        os.path.join(modules_dir, "minimalv1module"),
        str(tmpdir.join(handler_module_name)),
        new_name=handler_module_name,
        new_content_init_py=textwrap.dedent(
            """
            from inmanta.execute.proxy import DynamicProxy
            from inmanta.export import Exporter
            from inmanta.resources import Resource, resource

            @resource("__config__::Host", agent="agent.agentname", id_attribute="full_id")
            class Host(Resource):
                fields = ("network_id", "host_id", "full_id",)

                def get_network_id(exporter: Exporter, obj: DynamicProxy) -> int:
                    return obj.network.id

                def get_host_id(exporter: Exporter, obj: DynamicProxy) -> int:
                    return obj.id

                def get_full_id(exporter: Exporter, obj: DynamicProxy) -> tuple[int, str]:
                    return (Host.get_network_id(exporter, obj), Host.get_host_id(exporter, obj))
            """.strip(
                "\n"
            )
        ),
    )

    def setup_model(base: str) -> None:
        """
        Add handlers for the base model and set up the snippetcompiler.
        """
        handlers_addition: str = f"""
            import {handler_module_name} as handler

            # add dummy agent attribute for the handler
            Host.agent [1] -- std::AgentConfig
            implementation bind_agent for Host:
                self.agent = std::AgentConfig[agentname="host_agent"]
            end
            implement Host using bind_agent
        """.strip()
        full_model: str = "\n".join((base, handlers_addition))
        snippetcompiler.setup_for_snippet(full_model, add_to_module_path=[str(tmpdir)])

    async def get_hosts_by_network(version: int) -> abc.Mapping[int, abc.Set[int]]:
        resources: abc.Sequence[data.Resource] = await data.Resource.get_resources_for_version(env_id, version)
        hosts_by_network: dict[int, set[int]] = defaultdict(set)
        for resource in resources:
            if resource.resource_type == model.ResourceType("__config__::Host"):
                hosts_by_network[resource.attributes["network_id"]].add(resource.attributes["host_id"])
        return hosts_by_network

    # initial export
    with open(os.path.join(snippets_dir, "basic_example_full.cf")) as fd:
        setup_model(fd.read())
    version, _ = await snippetcompiler.do_export_and_deploy()
    hosts_by_network_full: abc.Mapping[int, abc.Set[int]] = await get_hosts_by_network(version)
    assert len(hosts_by_network_full) == 1000
    for network in hosts_by_network_full:
        assert len(hosts_by_network_full[network]) == 5

    # partial export: verify that only the example's set has changed
    with open(os.path.join(snippets_dir, "basic_example_partial.cf")) as fd:
        setup_model(fd.read())
    version, _ = await snippetcompiler.do_export_and_deploy(partial_compile=True)
    hosts_by_network_partial: abc.Mapping[int, abc.Set[int]] = await get_hosts_by_network(version)
    assert len(hosts_by_network_partial) == 1000
    for network in hosts_by_network_partial:
        assert len(hosts_by_network_partial[network]) == (1 if network == 0 else 5)

    # full equivalent export: verify that it is indeed equivalent
    with open(os.path.join(snippets_dir, "basic_example_full_result.cf")) as fd:
        setup_model(fd.read())
    version, _ = await snippetcompiler.do_export_and_deploy()
    hosts_by_network_equivalent: abc.Mapping[int, abc.Set[int]] = await get_hosts_by_network(version)
    assert hosts_by_network_equivalent == hosts_by_network_partial



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com


    Command line development guidelines
    ###################################

    do's and don'ts
    ----------------
    MUST NOT: sys.exit => use command.CLIException
    SHOULD NOT: print( => use logger for messages, only print for final output


    Entry points
    ------------
    @command annotation to register new command
"""
import argparse
import asyncio
import json
import logging
import os
import signal
import socket
import sys
import threading
import time
import traceback
from argparse import ArgumentParser
from asyncio import ensure_future
from collections import abc
from configparser import ConfigParser
from threading import Timer
from types import FrameType
from typing import Any, Callable, Coroutine, Dict, Optional

from tornado import gen
from tornado.ioloop import IOLoop
from tornado.util import TimeoutError

import inmanta.compiler as compiler
from inmanta import const, module, moduletool, protocol, util
from inmanta.ast import CompilerException, Namespace
from inmanta.ast import type as inmanta_type
from inmanta.command import CLIException, Commander, ShowUsageException, command
from inmanta.compiler import do_compile
from inmanta.config import Config, Option
from inmanta.const import EXIT_START_FAILED
from inmanta.export import cfg_env
from inmanta.logging import InmantaLoggerConfig, _is_on_tty
from inmanta.server.bootloader import InmantaBootloader
from inmanta.util import get_compiler_version
from inmanta.warnings import WarningsManager

try:
    import rpdb
except ImportError:
    rpdb = None

LOGGER = logging.getLogger("inmanta")


@command("server", help_msg="Start the inmanta server")
def start_server(options: argparse.Namespace) -> None:
    if options.config_file and not os.path.exists(options.config_file):
        LOGGER.warning("Config file %s doesn't exist", options.config_file)

    if options.config_dir and not os.path.isdir(options.config_dir):
        LOGGER.warning("Config directory %s doesn't exist", options.config_dir)

    util.ensure_event_loop()

    ibl = InmantaBootloader()
    setup_signal_handlers(ibl.stop)

    ioloop = IOLoop.current()

    # handle startup exceptions
    def _handle_startup_done(fut: asyncio.Future) -> None:
        if fut.cancelled():
            safe_shutdown(ioloop, ibl.stop)
        else:
            exc = fut.exception()
            if exc is not None:
                LOGGER.exception("Server setup failed", exc_info=exc)
                traceback.print_exception(type(exc), exc, exc.__traceback__)
                safe_shutdown(ioloop, ibl.stop)
            else:
                LOGGER.info("Server startup complete")

    ensure_future(ibl.start()).add_done_callback(_handle_startup_done)

    ioloop.start()
    LOGGER.info("Server shutdown complete")
    if not ibl.started:
        exit(EXIT_START_FAILED)


@command("agent", help_msg="Start the inmanta agent")
def start_agent(options: argparse.Namespace) -> None:
    from inmanta.agent import agent

    util.ensure_event_loop()

    a = agent.Agent()
    setup_signal_handlers(a.stop)
    IOLoop.current().add_callback(a.start)
    IOLoop.current().start()
    LOGGER.info("Agent Shutdown complete")


def dump_threads() -> None:
    print("----- Thread Dump ----")
    for th in threading.enumerate():
        print("---", th)
        if th.ident:
            traceback.print_stack(sys._current_frames()[th.ident], file=sys.stdout)
        print()
    sys.stdout.flush()


async def dump_ioloop_running() -> None:
    # dump async IO
    print("----- Async IO tasks ----")
    for task in asyncio.all_tasks():
        print(task)
    print()
    sys.stdout.flush()


def context_dump(ioloop: IOLoop) -> None:
    dump_threads()
    if hasattr(asyncio, "all_tasks"):
        ioloop.add_callback_from_signal(dump_ioloop_running)


def setup_signal_handlers(shutdown_function: Callable[[], Coroutine[Any, Any, None]]) -> None:
    """
    Make sure that shutdown_function is called when a SIGTERM or a SIGINT interrupt occurs.

    :param shutdown_function: The function that contains the shutdown logic.
    """
    # ensure correct ioloop
    ioloop = IOLoop.current()

    def hard_exit() -> None:
        context_dump(ioloop)
        sys.stdout.flush()
        # Hard exit, not sys.exit
        # ensure shutdown when the ioloop is stuck
        os._exit(const.EXIT_HARD)

    def handle_signal(signum: signal.Signals, frame: Optional[FrameType]) -> None:
        # force shutdown, even when the ioloop is stuck
        # schedule off the loop
        t = Timer(const.SHUTDOWN_GRACE_HARD, hard_exit)
        t.daemon = True
        t.start()
        ioloop.add_callback_from_signal(safe_shutdown_wrapper, shutdown_function)

    def handle_signal_dump(signum: signal.Signals, frame: Optional[FrameType]) -> None:
        context_dump(ioloop)

    signal.signal(signal.SIGTERM, handle_signal)
    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGUSR1, handle_signal_dump)
    if rpdb:
        rpdb.handle_trap()


def safe_shutdown(ioloop: IOLoop, shutdown_function: Callable[[], None]) -> None:
    def hard_exit() -> None:
        context_dump(ioloop)
        sys.stdout.flush()
        # Hard exit, not sys.exit
        # ensure shutdown when the ioloop is stuck
        os._exit(const.EXIT_HARD)

    # force shutdown, even when the ioloop is stuck
    # schedule off the loop
    t = Timer(const.SHUTDOWN_GRACE_HARD, hard_exit)
    t.daemon = True
    t.start()
    ioloop.add_callback(safe_shutdown_wrapper, shutdown_function)


async def safe_shutdown_wrapper(shutdown_function: Callable[[], Coroutine[Any, Any, None]]) -> None:
    """
    Wait 10 seconds to gracefully shutdown the instance.
    Afterwards stop the IOLoop
    Wait for 3 seconds to force stop
    """
    future = shutdown_function()
    try:
        timeout = IOLoop.current().time() + const.SHUTDOWN_GRACE_IOLOOP
        await gen.with_timeout(timeout, future)
    except TimeoutError:
        pass
    finally:
        IOLoop.current().stop()


class ExperimentalFeatureFlags:
    """
    Class to expose feature flag configs as options in a uniform matter
    """

    def __init__(self) -> None:
        self.metavar_to_option: Dict[str, Option[bool]] = {}

    def _get_name(self, option: Option[bool]) -> str:
        return f"flag_{option.name}"

    def add(self, option: Option[bool]) -> None:
        """Add an option to the set of feature flags"""
        self.metavar_to_option[self._get_name(option)] = option

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        """Add all feature flag options to the argument parser"""
        for metavar, option in self.metavar_to_option.items():
            parser.add_argument(
                f"--experimental-{option.name}",
                dest=metavar,
                help=option.documentation,
                action="store_true",
                default=False,
            )

    def read_options_to_config(self, options: argparse.Namespace) -> None:
        """
        This method takes input from the commandline parser
        and sets the appropriate feature flag config based
        on the parsed command line arguments

        :param options: the options, as parsed by argparse.
        """
        for metavar, option in self.metavar_to_option.items():
            value = getattr(options, metavar, False)
            if value:
                option.set("true")


def compiler_config(parser: argparse.ArgumentParser, parent_parsers: abc.Sequence[ArgumentParser]) -> None:
    """
    Configure the compiler of the export function
    """
    parser.add_argument("-e", dest="environment", help="The environment to compile this model for")
    parser.add_argument(
        "-X",
        "--extended-errors",
        dest="errors",
        help="Show stack traces for compile errors",
        action="store_true",
        default=argparse.SUPPRESS,
    )
    parser.add_argument("--server_address", dest="server", help="The address of the server hosting the environment")
    parser.add_argument("--server_port", dest="port", help="The port of the server hosting the environment")
    parser.add_argument("--username", dest="user", help="The username of the server")
    parser.add_argument("--password", dest="password", help="The password of the server")
    parser.add_argument("--ssl", help="Enable SSL", action="store_true", default=False)
    parser.add_argument("--ssl-ca-cert", dest="ca_cert", help="Certificate authority for SSL")
    parser.add_argument(
        "--export-compile-data",
        dest="export_compile_data",
        help="Export structured json containing compile data such as occurred errors.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--export-compile-data-file",
        dest="export_compile_data_file",
        help="File to export compile data to. If omitted %s is used." % compiler.config.default_compile_data_file,
    )
    parser.add_argument(
        "--no-cache",
        dest="feature_compiler_cache",
        help="Disable caching of compiled CF files",
        action="store_false",
        default=True,
    )
    parser.add_argument(
        "--experimental-data-trace",
        dest="datatrace",
        help="Experimental data trace tool useful for debugging",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--experimental-dataflow-graphic",
        dest="dataflow_graphic",
        help="Experimental graphic data flow visualization",
        action="store_true",
        default=False,
    )

    parser.add_argument("-f", dest="main_file", help="Main file", default="main.cf")
    moduletool.add_deps_check_arguments(parser)


@command(
    "compile", help_msg="Compile the project to a configuration model", parser_config=compiler_config, require_project=True
)
def compile_project(options: argparse.Namespace) -> None:
    if options.environment is not None:
        Config.set("config", "environment", options.environment)

    if options.server is not None:
        Config.set("compiler_rest_transport", "host", options.server)

    if options.port is not None:
        Config.set("compiler_rest_transport", "port", options.port)

    if options.user is not None:
        Config.set("compiler_rest_transport", "username", options.user)

    if options.password is not None:
        Config.set("compiler_rest_transport", "password", options.password)

    if options.ssl:
        Config.set("compiler_rest_transport", "ssl", "true")

    if options.ca_cert is not None:
        Config.set("compiler_rest_transport", "ssl-ca-cert-file", options.ca_cert)

    if options.export_compile_data is True:
        Config.set("compiler", "export_compile_data", "true")

    if options.export_compile_data_file is not None:
        Config.set("compiler", "export_compile_data_file", options.export_compile_data_file)

    if options.feature_compiler_cache is False:
        Config.set("compiler", "cache", "false")

    if options.datatrace is True:
        Config.set("compiler", "datatrace_enable", "true")

    if options.dataflow_graphic is True:
        Config.set("compiler", "dataflow_graphic_enable", "true")

    strict_deps_check = moduletool.get_strict_deps_check(
        no_strict_deps_check=options.no_strict_deps_check, strict_deps_check=options.strict_deps_check
    )
    module.Project.get(options.main_file, strict_deps_check=strict_deps_check)

    if options.profile:
        import cProfile
        import pstats

        cProfile.runctx("do_compile()", globals(), {}, "run.profile")
        p = pstats.Stats("run.profile")
        p.strip_dirs().sort_stats("time").print_stats(20)
    else:
        t1 = time.time()
        do_compile()
        LOGGER.debug("Compile time: %0.03f seconds", time.time() - t1)


@command("list-commands", help_msg="Print out an overview of all commands", add_verbose_flag=False)
def list_commands(options: argparse.Namespace) -> None:
    print("The following commands are available:")
    for cmd, info in Commander.commands().items():
        print(" %s: %s" % (cmd, info["help"]))


def help_parser_config(parser: argparse.ArgumentParser, parent_parsers: abc.Sequence[ArgumentParser]) -> None:
    parser.add_argument("subcommand", help="Output help for a particular subcommand", nargs="?", default=None)


@command("help", help_msg="show a help message and exit", parser_config=help_parser_config, add_verbose_flag=False)
def help_command(options: argparse.Namespace) -> None:
    if options.subcommand is None:
        cmd_parser().print_help()
    else:
        subc = options.subcommand
        parser = cmd_parser()
        parser.parse_args([subc, "-h"])
    sys.exit(0)


@command(
    "modules",
    help_msg="Subcommand to manage modules",
    parser_config=moduletool.ModuleTool.modules_parser_config,
    aliases=["module"],
)
def modules(options: argparse.Namespace) -> None:
    tool = moduletool.ModuleTool()
    tool.execute(options.cmd, options)


@command("project", help_msg="Subcommand to manage the project", parser_config=moduletool.ProjectTool.parser_config)
def project(options: argparse.Namespace) -> None:
    tool = moduletool.ProjectTool()
    tool.execute(options.cmd, options)


def deploy_parser_config(parser: argparse.ArgumentParser, parent_parsers: abc.Sequence[ArgumentParser]) -> None:
    parser.add_argument("--dry-run", help="Only report changes", action="store_true", dest="dryrun")
    parser.add_argument("-f", dest="main_file", help="Main file", default="main.cf")


@command("deploy", help_msg="Deploy with a inmanta all-in-one setup", parser_config=deploy_parser_config, require_project=True)
def deploy(options: argparse.Namespace) -> None:
    module.Project.get(options.main_file)
    from inmanta import deploy as deploy_module

    run = deploy_module.Deploy(options)
    try:
        if not run.setup():
            raise Exception("Failed to setup an embedded Inmanta orchestrator.")
        run.run()
    finally:
        run.stop()


def export_parser_config(parser: argparse.ArgumentParser, parent_parsers: abc.Sequence[ArgumentParser]) -> None:
    """
    Configure the compiler of the export function
    """
    parser.add_argument("-g", dest="depgraph", help="Dump the dependency graph", action="store_true")
    parser.add_argument(
        "-j",
        dest="json",
        help="Do not submit to the server but only store the json that would have been " "submitted in the supplied file",
    )
    parser.add_argument("-e", dest="environment", help="The environment to compile this model for")
    parser.add_argument("-d", dest="deploy", help="Trigger a deploy for the exported version", action="store_true")
    parser.add_argument(
        "--full",
        dest="full_deploy",
        help="Make the agents execute a full deploy instead of an incremental deploy. "
        "Should be used together with the -d option",
        action="store_true",
        default=False,
    )
    parser.add_argument("-m", dest="model", help="Also export the complete model", action="store_true", default=False)
    parser.add_argument("--server_address", dest="server", help="The address of the server to submit the model to")
    parser.add_argument("--server_port", dest="port", help="The port of the server to submit the model to")
    parser.add_argument("--token", dest="token", help="The token to auth to the server")
    parser.add_argument("--ssl", help="Enable SSL", action=argparse.BooleanOptionalAction, default=None)
    parser.add_argument("--ssl-ca-cert", dest="ca_cert", help="Certificate authority for SSL")
    parser.add_argument(
        "-X",
        "--extended-errors",
        dest="errors",
        help="Show stack traces for compile errors",
        action="store_true",
        default=argparse.SUPPRESS,
    )
    parser.add_argument("-f", dest="main_file", help="Main file", default="main.cf")
    parser.add_argument(
        "--metadata",
        dest="metadata",
        help="JSON metadata why this compile happened. If a non-json string is "
        "passed it is used as the 'message' attribute in the metadata.",
        default=None,
    )
    parser.add_argument(
        "--model-export",
        dest="model_export",
        help="Export the configuration model to the server as metadata.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--export-plugin",
        dest="export_plugin",
        help="Only use this export plugin. This option also disables the execution of the plugins listed in "
        "the configuration file in the export setting.",
        default=None,
    )

    parser.add_argument(
        "--export-compile-data",
        dest="export_compile_data",
        help="Export structured json containing compile data such as occurred errors.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--export-compile-data-file",
        dest="export_compile_data_file",
        help="File to export compile data to. If omitted %s is used." % compiler.config.default_compile_data_file,
    )
    parser.add_argument(
        "--no-cache",
        dest="feature_compiler_cache",
        help="Disable caching of compiled CF files",
        action="store_false",
        default=True,
    )
    parser.add_argument(
        "--partial",
        dest="partial_compile",
        help=(
            "Execute a partial export. Does not upload new Python code to the server: it is assumed to be unchanged since the"
            " last full export. Multiple partial exports for disjunct resource sets may be performed concurrently but not"
            " concurrent with a full export. When used in combination with the `--json` option, 0 is used as a placeholder for"
            " the model version."
        ),
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--delete-resource-set",
        dest="delete_resource_set",
        help="Remove a resource set as part of a partial compile. This option can be provided multiple times and should always "
        "be used together with the --partial option.",
        action="append",
    )
    moduletool.add_deps_check_arguments(parser)


@command("export", help_msg="Export the configuration", parser_config=export_parser_config, require_project=True)
def export(options: argparse.Namespace) -> None:
    if not options.partial_compile and options.delete_resource_set:
        raise CLIException(
            "The --delete-resource-set option should always be used together with the --partial option", exitcode=1
        )
    if options.environment is not None:
        Config.set("config", "environment", options.environment)

    if options.server is not None:
        Config.set("compiler_rest_transport", "host", options.server)

    if options.port is not None:
        Config.set("compiler_rest_transport", "port", options.port)

    if options.token is not None:
        Config.set("compiler_rest_transport", "token", options.token)

    if options.ssl is not None:
        Config.set("compiler_rest_transport", "ssl", f"{options.ssl}".lower())

    if options.ca_cert is not None:
        Config.set("compiler_rest_transport", "ssl-ca-cert-file", options.ca_cert)

    if options.export_compile_data is True:
        Config.set("compiler", "export_compile_data", "true")

    if options.export_compile_data_file is not None:
        Config.set("compiler", "export_compile_data_file", options.export_compile_data_file)

    if options.feature_compiler_cache is False:
        Config.set("compiler", "cache", "false")

    # try to parse the metadata as json. If a normal string, create json for it.
    if options.metadata is not None and len(options.metadata) > 0:
        try:
            metadata = json.loads(options.metadata)
        except json.decoder.JSONDecodeError:
            metadata = {"message": options.metadata}
    else:
        metadata = {"message": "Manual compile on the CLI by user"}

    if "cli-user" not in metadata and "USERNAME" in os.environ:
        metadata["cli-user"] = os.environ["USERNAME"]

    if "hostname" not in metadata:
        metadata["hostname"] = socket.gethostname()

    if "type" not in metadata:
        metadata["type"] = "manual"

    strict_deps_check = moduletool.get_strict_deps_check(
        no_strict_deps_check=options.no_strict_deps_check, strict_deps_check=options.strict_deps_check
    )
    module.Project.get(options.main_file, strict_deps_check=strict_deps_check)

    from inmanta.export import Exporter  # noqa: H307

    exp = None
    types: Optional[Dict[str, inmanta_type.Type]]
    scopes: Optional[Namespace]
    try:
        (types, scopes) = do_compile()
    except Exception as e:
        exp = e
        types, scopes = (None, None)

    # Even if the compile failed we might have collected additional data such as unknowns. So
    # continue the export

    export = Exporter(options)
    results = export.run(
        types,
        scopes,
        metadata=metadata,
        model_export=options.model_export,
        export_plugin=options.export_plugin,
        partial_compile=options.partial_compile,
        resource_sets_to_remove=options.delete_resource_set,
    )
    version = results[0]

    if exp is not None:
        raise exp

    if options.deploy:
        conn = protocol.SyncClient("compiler")
        LOGGER.info("Triggering deploy for version %d" % version)
        tid = cfg_env.get()
        agent_trigger_method = const.AgentTriggerMethod.get_agent_trigger_method(options.full_deploy)
        conn.release_version(tid, version, True, agent_trigger_method)


def cmd_parser() -> argparse.ArgumentParser:
    # create the argument compiler

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", action="store_true", dest="profile", help="Profile this run of the program")
    parser.add_argument("-c", "--config", dest="config_file", help="Use this config file", default=None)
    parser.add_argument(
        "--config-dir",
        dest="config_dir",
        help="The directory containing the Inmanta configuration files",
        default="/etc/inmanta/inmanta.d",
    )
    parser.add_argument("--log-file", dest="log_file", help="Path to the logfile")
    parser.add_argument(
        "--log-file-level",
        dest="log_file_level",
        choices=["0", "1", "2", "3", "4", "ERROR", "WARNING", "INFO", "DEBUG", "TRACE"],
        default="INFO",
        help="Log level for messages going to the logfile: 0=ERROR, 1=WARNING, 2=INFO, 3=DEBUG",
    )
    parser.add_argument("--timed-logs", dest="timed", help="Add timestamps to logs", action="store_true")
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Log level for messages going to the console. Default is warnings,"
        "-v warning, -vv info, -vvv debug and -vvvv trace",
    )
    parser.add_argument(
        "--warnings",
        dest="warnings",
        choices=["warn", "ignore", "error"],
        default="warn",
        help="The warning behaviour. Must be one of 'warn', 'ignore', 'error'",
    )
    parser.add_argument(
        "-X", "--extended-errors", dest="errors", help="Show stack traces for errors", action="store_true", default=False
    )
    parser.add_argument(
        "--version",
        action="store_true",
        dest="inmanta_version",
        help="Show the version of the installed Inmanta product and the version of its subcomponents",
        default=False,
        required=False,
    )

    verbosity_parser = argparse.ArgumentParser(add_help=False)
    verbosity_parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=argparse.SUPPRESS,
        help="Log level for messages going to the console. Default is warnings,"
        "-v warning, -vv info, -vvv debug and -vvvv trace",
    )

    subparsers = parser.add_subparsers(title="commands")
    for cmd_name, cmd_options in Commander.commands().items():
        parent_parsers: list[argparse.ArgumentParser] = []
        if cmd_options["add_verbose_flag"]:
            parent_parsers.append(verbosity_parser)
        cmd_subparser = subparsers.add_parser(
            cmd_name, help=cmd_options["help"], aliases=cmd_options["aliases"], parents=parent_parsers
        )
        if cmd_options["parser_config"] is not None:
            cmd_options["parser_config"](cmd_subparser, parent_parsers)
        cmd_subparser.set_defaults(func=cmd_options["function"])
        cmd_subparser.set_defaults(require_project=cmd_options["require_project"])

    return parser


def print_versions_installed_components_and_exit() -> None:
    # coroutine to make sure event loop is running for server slices
    async def print_status() -> None:
        bootloader = InmantaBootloader()
        app_context = bootloader.load_slices()
        product_metadata = app_context.get_product_metadata()
        extension_statuses = app_context.get_extension_statuses()

        if product_metadata.version:
            print(f"{product_metadata.product} ({product_metadata.edition}): {product_metadata.version}")
        else:
            print(f"{product_metadata.product} ({product_metadata.edition}): version unknown")
        print(f"Compiler version: {get_compiler_version()}")
        if extension_statuses:
            print("Extensions:")
            for ext_status in extension_statuses:
                print(f"    * {ext_status.name}: {ext_status.version}")
        else:
            print("Extensions: No extensions found")

        await bootloader.stop()

    asyncio.run(print_status())
    sys.exit(0)


def app() -> None:
    """
    Run the compiler
    """
    log_config = InmantaLoggerConfig.get_instance()

    # do an initial load of known config files to build the libdir path
    Config.load_config()
    parser = cmd_parser()
    options, other = parser.parse_known_args()
    options.other = other

    log_config.apply_options(options)

    logging.captureWarnings(True)

    if options.config_file and not os.path.exists(options.config_file):
        LOGGER.warning("Config file %s doesn't exist", options.config_file)

    # Load the configuration
    Config.load_config(options.config_file, options.config_dir)

    if options.inmanta_version:
        print_versions_installed_components_and_exit()

    if options.warnings is not None:
        Config.set("warnings", "default", options.warnings)

    config = Config.get()
    assert isinstance(config, ConfigParser)
    WarningsManager.apply_config(config["warnings"] if "warnings" in config else None)

    # start the command
    if not hasattr(options, "func"):
        # show help
        parser.print_usage()
        return

    def report(e: BaseException) -> None:
        if not options.errors:
            if isinstance(e, CompilerException):
                print(e.format_trace(indent="  "), file=sys.stderr)
            else:
                print(str(e), file=sys.stderr)
        else:
            sys.excepthook(*sys.exc_info())

        if isinstance(e, CompilerException):
            from inmanta.compiler.help.explainer import ExplainerFactory

            helpmsg = ExplainerFactory().explain_and_format(e, plain=not _is_on_tty())
            if helpmsg is not None:
                print(helpmsg)

    try:
        options.func(options)
    except ShowUsageException as e:
        print(e.args[0], file=sys.stderr)
        parser.print_usage()
    except CLIException as e:
        report(e)
        sys.exit(e.exitcode)
    except Exception as e:
        report(e)
        sys.exit(1)
    except KeyboardInterrupt as e:
        report(e)
        sys.exit(1)
    sys.exit(0)


if __name__ == "__main__":
    app()



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
from collections import abc
from typing import Dict, Generic, List, Optional, TypeVar

import inmanta.execute.dataflow as dataflow
from inmanta.ast import LocatableString, Location, NotFoundException, OptionalValueException, Range, RuntimeException
from inmanta.ast.statements import (
    AssignStatement,
    AttributeAssignmentLHS,
    ExpressionStatement,
    RawResumer,
    Statement,
    VariableReferenceHook,
    VariableResumer,
)
from inmanta.ast.statements.assign import Assign, SetAttribute
from inmanta.execute.dataflow import DataflowGraph
from inmanta.execute.runtime import (
    QueueScheduler,
    RawUnit,
    Resolver,
    ResultCollector,
    ResultVariable,
    ResultVariableProxy,
    VariableABC,
)
from inmanta.execute.util import NoneValue
from inmanta.parser import ParserException
from inmanta.stable_api import stable_api

LOGGER = logging.getLogger(__name__)


R = TypeVar("R", bound="Reference")


@stable_api
class Reference(ExpressionStatement):
    """
    This class represents a reference to a value

    :ivar name: The name of the Reference as a string.
    """

    __slots__ = ("locatable_name", "name", "full_name")

    def __init__(self, name: LocatableString) -> None:
        ExpressionStatement.__init__(self)
        self.locatable_name = name
        self.name = str(name)
        self.full_name = str(name)

    def normalize(self, *, lhs_attribute: Optional[AttributeAssignmentLHS] = None) -> None:
        split: abc.Sequence[str] = self.name.rsplit("::", maxsplit=1)
        if len(split) > 1:
            # fail-fast if namespace does not exist
            try:
                self.namespace.lookup_namespace(split[0])
            except NotFoundException as e:
                e.set_statement(self)
                raise

    def requires(self) -> List[str]:
        return [self.full_name]

    def requires_emit(
        self, resolver: Resolver, queue: QueueScheduler, *, propagate_unset: bool = False
    ) -> Dict[object, VariableABC]:
        requires: Dict[object, VariableABC] = super().requires_emit(resolver, queue)
        # FIXME: may be done more efficient?
        requires[self.name] = resolver.lookup(self.full_name)
        return requires

    def requires_emit_gradual(
        self, resolver: Resolver, queue: QueueScheduler, resultcollector: ResultCollector, *, propagate_unset: bool = False
    ) -> Dict[object, VariableABC]:
        requires: Dict[object, VariableABC] = self._requires_emit_promises(resolver, queue)
        var: ResultVariable = resolver.lookup(self.full_name)
        var.listener(resultcollector, self.location)
        requires[self.name] = var
        return requires

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        super().execute(requires, resolver, queue)
        return requires[self.name]

    def execute_direct(self, requires: abc.Mapping[str, object]) -> object:
        if self.name not in requires:
            raise NotFoundException(self, "Could not resolve the value %s in this static context" % self.name)
        return requires[self.name]

    def get_root_variable(self) -> "Reference":
        """
        Returns the root reference node. e.g. for a.b.c.d, returns the reference for a.
        """
        return self

    def fully_qualified(self: R) -> R:
        """
        If this reference is already fully qualified, returns it unchanged. Otherwise, returns a new fully qualified reference
        to this name in this reference's namespace.
        This fully qualified reference is not guaranteed to resolve to the same object. It is the caller's responsibility to
        only request a fully qualified name when appropriate.
        """
        if "::" in self.name:
            return self
        fully_qualified_name: str = f"{self.namespace.name}::{self.name}"
        locatable: LocatableString = LocatableString(
            fully_qualified_name, Range("__internal__", 1, 1, 1, 1), -1, self.namespace
        )
        result: R = self.__class__(locatable)
        result.location = locatable.location
        return result

    def as_assign(self, value: ExpressionStatement, list_only: bool = False) -> AssignStatement:
        if list_only:
            raise ParserException(self.location, "+=", "Can not perform += on variable %s" % self.name)
        return Assign(self.locatable_name, value)

    def root_in_self(self) -> "Reference":
        if self.name == "self":
            return self
        else:
            ref = Reference("self")
            self.copy_location(ref)
            attr_ref = AttributeReference(ref, self.locatable_name)
            self.copy_location(attr_ref)
            return attr_ref

    def get_dataflow_node(self, graph: DataflowGraph) -> dataflow.AssignableNodeReference:
        return graph.resolver.get_dataflow_node(self.name)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.name


T = TypeVar("T")


class VariableReader(VariableResumer, Generic[T]):
    """
    Resumes execution on a variable when it becomes avaiable, then connects the target proxy variable to it.
    Optionally subscribes a result collector to intermediate values.
    """

    __slots__ = ("owner", "target", "resultcollector")

    def __init__(self, owner: Statement, target: ResultVariableProxy[T], resultcollector: Optional[ResultCollector[T]]) -> None:
        super().__init__()
        self.owner: Statement = owner
        self.target: ResultVariableProxy[T] = target
        self.resultcollector: Optional[ResultCollector[T]] = resultcollector

    def variable_resume(
        self,
        variable: VariableABC[T],
        resolver: Resolver,
        queue_scheduler: QueueScheduler,
    ) -> None:
        if self.resultcollector:
            variable.listener(self.resultcollector, self.owner.location)
        self.target.connect(variable)


class IsDefinedGradual(VariableResumer, RawResumer, ResultCollector[object]):
    """
    Fill target variable with is defined result as soon as it gets known.
    """

    __slots__ = ("owner", "target")

    def __init__(self, owner: Statement, target: ResultVariable[bool]) -> None:
        super().__init__()
        self.owner: Statement = owner
        self.target: ResultVariable[bool] = target

    def pure_gradual(self) -> bool:
        # freezing an empty variable causes progress
        return False

    def receive_result(self, value: object, location: Location) -> bool:
        """
        Gradually receive an assignment to the referenced variable. Sets the target variable to True because to receive a single
        value implies that the variable is defined.
        """
        self.target.set_value(True, self.owner.location)
        return True

    def variable_resume(
        self,
        variable: VariableABC[object],
        resolver: Resolver,
        queue_scheduler: QueueScheduler,
    ) -> None:
        if variable.is_ready():
            self.target.set_value(self._target_value(variable), self.owner.location)
        else:
            # gradual execution: as soon as a value comes in, the result is known
            variable.listener(self, self.owner.location)
            # wait for variable completeness in case no value comes in at all
            RawUnit(queue_scheduler, resolver, {self: variable}, self, override_exception_location=False)

    def resume(self, requires: Dict[object, VariableABC], resolver: Resolver, queue_scheduler: QueueScheduler) -> None:
        self.target.set_value(self._target_value(requires[self]), self.owner.location)

    def _target_value(self, variable: VariableABC[object]) -> bool:
        """
        Returns the target value based on the attribute variable's value or absence of a value.
        """
        try:
            value = variable.get_value()
            if isinstance(value, list):
                return len(value) != 0
            elif isinstance(value, NoneValue):
                return False
            return True
        except OptionalValueException:
            return False

    def emit(self, resolver: Resolver, queue: QueueScheduler) -> None:
        raise RuntimeException(self, "%s is not an actual AST node, it should never be executed" % self.__class__.__name__)

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        raise RuntimeException(self, "%s is not an actual AST node, it should never be executed" % self.__class__.__name__)


class AttributeReference(Reference):
    """
    This variable refers to an attribute. This is mostly used to refer to
    attributes of a class or class instance.
    """

    __slots__ = ("attribute", "instance")

    def __init__(self, instance: Reference, attribute: LocatableString) -> None:
        range: Range = Range(
            instance.locatable_name.location.file,
            instance.locatable_name.lnr,
            instance.locatable_name.start,
            attribute.elnr,
            attribute.end,
        )
        reference: LocatableString = LocatableString(
            "%s.%s" % (instance.full_name, attribute), range, instance.locatable_name.lexpos, instance.namespace
        )
        Reference.__init__(self, reference)
        self.attribute = attribute

        # a reference to the instance
        self.instance = instance

    def requires(self) -> List[str]:
        return self.instance.requires()

    def requires_emit(
        self, resolver: Resolver, queue: QueueScheduler, *, propagate_unset: bool = False
    ) -> Dict[object, VariableABC]:
        return self.requires_emit_gradual(resolver, queue, None, propagate_unset=propagate_unset)

    def requires_emit_gradual(
        self,
        resolver: Resolver,
        queue: QueueScheduler,
        resultcollector: Optional[ResultCollector],
        *,
        propagate_unset: bool = False,
    ) -> Dict[object, VariableABC]:
        requires: Dict[object, VariableABC] = self._requires_emit_promises(resolver, queue)

        # The tricky one!

        # introduce temp variable to contain the eventual result of this stmt
        temp = ResultVariableProxy()
        # construct waiter
        reader: VariableReader = VariableReader(owner=self, target=temp, resultcollector=resultcollector)
        hook: VariableReferenceHook = VariableReferenceHook(
            self.instance,
            str(self.attribute),
            variable_resumer=reader,
            propagate_unset=propagate_unset,
        )
        self.copy_location(hook)
        hook.schedule(resolver, queue)
        # wait for the attribute value
        requires[self] = temp

        return requires

    def execute(self, requires: Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        ExpressionStatement.execute(self, requires, resolver, queue)
        # helper returned: return result
        return requires[self]

    def get_root_variable(self) -> "Reference":
        """
        Returns the root reference node. e.g. for a.b.c.d, returns the reference for a.
        """
        return self.instance.get_root_variable()

    def as_assign(self, value: ExpressionStatement, list_only: bool = False) -> AssignStatement:
        return SetAttribute(self.instance, str(self.attribute), value, list_only)

    def root_in_self(self) -> Reference:
        out = AttributeReference(self.instance.root_in_self(), str(self.attribute))
        self.copy_location(out)
        return out

    def get_dataflow_node(self, graph: DataflowGraph) -> dataflow.AttributeNodeReference:
        assert self.instance is not None
        return dataflow.AttributeNodeReference(self.instance.get_dataflow_node(graph), str(self.attribute))

    def __repr__(self) -> str:
        return "%s.%s" % (repr(self.instance), str(self.attribute))



"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import json
import logging
import os
import uuid
from typing import TYPE_CHECKING, Dict, List, Optional, Union, cast

from tornado import routing, web

from inmanta import data
from inmanta.const import ApiDocsFormat
from inmanta.data.model import FeatureStatus, SliceStatus, StatusResponse
from inmanta.protocol import exceptions, handle, methods, methods_v2
from inmanta.protocol.common import HTML_CONTENT_WITH_UTF8_CHARSET, ReturnValue, attach_warnings
from inmanta.protocol.openapi.converter import OpenApiConverter
from inmanta.protocol.openapi.model import OpenAPI
from inmanta.server import SLICE_COMPILER, SLICE_DATABASE, SLICE_SERVER, SLICE_TRANSPORT
from inmanta.server import config as opt
from inmanta.server import protocol
from inmanta.types import Apireturn, JsonType, Warnings

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from inmanta.server.services.compilerservice import CompilerService

DBLIMIT = 100000


class Server(protocol.ServerSlice):
    """
    The central Inmanta server that communicates with clients and agents and persists configuration
    information
    """

    _server_storage: Dict[str, str]
    compiler: "CompilerService"
    _server: protocol.Server

    def __init__(self) -> None:
        super().__init__(name=SLICE_SERVER)
        LOGGER.info("Starting server endpoint")

    def get_dependencies(self) -> List[str]:
        return [SLICE_DATABASE, SLICE_COMPILER]

    def get_depended_by(self) -> List[str]:
        return [SLICE_TRANSPORT]

    async def prestart(self, server: protocol.Server) -> None:
        self._server = server
        self._server_storage: Dict[str, str] = self.check_storage()
        self.compiler: "CompilerService" = cast("CompilerService", server.get_slice(SLICE_COMPILER))
        self._handlers.append(routing.Rule(routing.PathMatches(r"/dashboard"), web.RedirectHandler, dict(url=r"/console")))
        self._handlers.append(
            routing.Rule(routing.PathMatches(r"/dashboard/(.*)"), web.RedirectHandler, dict(url=r"/console/{0}"))
        )

    def check_storage(self) -> Dict[str, str]:
        """
        Check if the server storage is configured and ready to use.
        """

        def _ensure_directory_exist(directory: str, *subdirs: str) -> str:
            directory = os.path.join(directory, *subdirs)
            if not os.path.exists(directory):
                os.mkdir(directory)
            return directory

        state_dir = opt.state_dir.get()
        server_state_dir = os.path.join(state_dir, "server")
        dir_map = {"server": _ensure_directory_exist(state_dir, "server")}
        dir_map["files"] = _ensure_directory_exist(server_state_dir, "files")
        dir_map["environments"] = _ensure_directory_exist(server_state_dir, "environments")
        dir_map["agents"] = _ensure_directory_exist(server_state_dir, "agents")
        dir_map["logs"] = _ensure_directory_exist(opt.log_dir.get())
        return dir_map

    @handle(methods.notify_change_get, env="id")
    async def notify_change_get(self, env: data.Environment, update: bool) -> Apireturn:
        result = await self.notify_change(env, update, {})
        return result

    @handle(methods.notify_change, env="id")
    async def notify_change(self, env: data.Environment, update: bool, metadata: JsonType) -> Apireturn:
        LOGGER.info("Received change notification for environment %s", env.id)
        if "type" not in metadata:
            metadata["type"] = "api"

        if "message" not in metadata:
            metadata["message"] = "Recompile trigger through API call"

        warnings = await self._async_recompile(env, update, metadata=metadata)

        return attach_warnings(200, None, warnings)

    async def _async_recompile(self, env: data.Environment, update_repo: bool, metadata: JsonType = {}) -> Warnings:
        """
        Recompile an environment in a different thread and taking wait time into account.
        """
        _, warnings = await self.compiler.request_recompile(
            env=env, force_update=update_repo, do_export=True, remote_id=uuid.uuid4(), metadata=metadata
        )
        return warnings

    @handle(methods.get_server_status)
    async def get_server_status(self) -> StatusResponse:
        product_metadata = self.feature_manager.get_product_metadata()
        if product_metadata.version is None:
            raise exceptions.ServerError(
                "Could not find version number for the inmanta compiler."
                "Is inmanta installed? Use setuptools install or setuptools dev to install."
            )

        async def collect_for_slice(slice_name: str, slice: protocol.ServerSlice) -> SliceStatus:
            try:
                return SliceStatus(name=slice_name, status=await asyncio.wait_for(slice.get_status(), 0.1))
            except asyncio.TimeoutError:
                return SliceStatus(
                    name=slice_name,
                    status={
                        "error": f"timeout on data collection for {slice_name}, "
                        "consult the server log for additional information"
                    },
                )
            except Exception:
                LOGGER.error(
                    f"The following error occured while trying to determine the status of slice {slice_name}",
                    exc_info=True,
                )
                return SliceStatus(name=slice_name, status={"error": "An unexpected error occurred, reported to server log"})

        slices = await asyncio.gather(
            *(collect_for_slice(slice_name, slice) for slice_name, slice in self._server.get_slices().items())
        )

        response = StatusResponse(
            product=product_metadata.product,
            edition=product_metadata.edition,
            version=product_metadata.version,
            license=product_metadata.license,
            extensions=self.get_extension_statuses(list(self._server.get_slices().values())),
            slices=slices,
            features=[
                FeatureStatus(slice=feature.slice, name=feature.name, value=self.feature_manager.get_value(feature))
                for feature in self.feature_manager.get_features()
            ],
        )

        return response

    @handle(methods_v2.get_api_docs)
    async def get_api_docs(self, format: Optional[ApiDocsFormat] = ApiDocsFormat.swagger) -> ReturnValue[Union[OpenAPI, str]]:
        url_map = self._server._transport.get_global_url_map(self._server.get_slices().values())
        feature_manager = self.feature_manager
        openapi = OpenApiConverter(url_map, feature_manager)
        # Get rid of none values with custom json encoder
        openapi_json_str = openapi.generate_openapi_json()
        if format == ApiDocsFormat.openapi:
            openapi_dict = json.loads(openapi_json_str)
            return ReturnValue(response=openapi_dict)

        swagger_html = openapi.get_swagger_html(openapi_json_str)
        return ReturnValue(content_type=HTML_CONTENT_WITH_UTF8_CHARSET, response=swagger_html)



"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import json
from typing import Dict, List, Optional, Set

from inmanta import resources
from inmanta.data.model import AttributeDiff, ResourceDiff, ResourceDiffStatus, ResourceIdStr


class Attribute:
    def __init__(self, name: str, value: object) -> None:
        self._name = name
        self._value = value

        if self._name == "requires":
            # Sort the requires list
            self._value = sorted([resources.Id.parse_id(req).resource_str() for req in self._value])

        self._compare_value: Optional[str] = None

    @property
    def value(self) -> object:
        """The value of the attribute"""
        return self._value

    @property
    def compare_value(self) -> str:
        """The string representation of the value, which can be used for comparison"""
        self._generate_compare_value()

        return self._compare_value

    def _generate_compare_value(self) -> None:
        """Generate a value that can be used for comparison"""
        if self._compare_value is not None:
            return

        if self.value is None:
            self._compare_value = ""
        elif isinstance(self.value, (dict, list)):
            self._compare_value = json.dumps(self.value, indent=4, sort_keys=True)
        else:
            self._compare_value = str(self.value)

    def compare(self, other: "Attribute") -> Optional[AttributeDiff]:
        """Compare this value with other. Other is considered the original value"""
        if self.compare_value == other.compare_value:
            return None

        diff = AttributeDiff(
            from_value=other.value,
            to_value=self.value,
            from_value_compare=other.compare_value,
            to_value_compare=self.compare_value,
        )
        return diff

    def added(self) -> AttributeDiff:
        """Return an attribute diff as if this attribute is newly added"""
        return AttributeDiff(
            from_value=None,
            to_value=self.value,
            from_value_compare="",
            to_value_compare=self.compare_value,
        )

    def removed(self) -> AttributeDiff:
        """Return an attribute diff as if this attribute is removed in a later version"""
        return AttributeDiff(
            from_value=self.value,
            to_value=None,
            from_value_compare=self.compare_value,
            to_value_compare="",
        )


class Resource:
    def __init__(self, resource_id: ResourceIdStr, attributes: Dict[str, object]) -> None:
        self.resource_id = resource_id
        self._attributes = {name: Attribute(name, value) for name, value in attributes.items() if name != "version"}

    def compare(self, other: "Resource") -> Optional[ResourceDiff]:
        """Compare this resource with another: check which attributes are added, modified and removed.
        The other resource is considered to be the original"""
        other_attributes = set(other._attributes.keys())
        our_attributes = set(self._attributes.keys())

        diff = {}
        # Generate the diff for each attribute
        for added in our_attributes - other_attributes:
            diff[added] = self._attributes[added].added()

        for removed in other_attributes - our_attributes:
            diff[removed] = other._attributes[removed].removed()

        for name in other_attributes.intersection(our_attributes):
            attr_diff = self._attributes[name].compare(other._attributes[name])
            if attr_diff is not None:
                diff[name] = attr_diff

        if diff:
            return ResourceDiff(resource_id=self.resource_id, attributes=diff, status=ResourceDiffStatus.modified)

        return None

    def added(self) -> ResourceDiff:
        """Return a diff as if this resource is newly added"""
        return ResourceDiff(
            resource_id=self.resource_id,
            attributes={name: attr.added() for name, attr in self._attributes.items()},
            status=ResourceDiffStatus.added,
        )

    def removed(self) -> ResourceDiff:
        """Return a diff as if this resource is removed"""
        return ResourceDiff(
            resource_id=self.resource_id,
            attributes={name: attr.removed() for name, attr in self._attributes.items()},
            status=ResourceDiffStatus.deleted,
        )

    def unmodified(self) -> ResourceDiff:
        """Return a diff as if this resource is not modified"""
        return ResourceDiff(
            resource_id=self.resource_id,
            attributes={},
            status=ResourceDiffStatus.unmodified,
        )


class Version:
    def __init__(self, resources: Dict[ResourceIdStr, Resource]) -> None:
        self._resources = resources

    def get_resource_set(self) -> Set[ResourceIdStr]:
        """The names of the resources in this version"""
        return set(self._resources.keys())

    def generate_diff(self, other: "Version", include_unmodified: bool = False) -> List[ResourceDiff]:
        """Compare this version with another: check which resources are added, removed and modified.
        The other version is considered to be the original."""
        our_set = self.get_resource_set()
        other_set = other.get_resource_set()
        result: List[ResourceDiff] = []

        added = list(our_set - other_set)
        removed = list(other_set - our_set)
        result.extend(self._resources[x].added() for x in added)
        result.extend(other._resources[x].removed() for x in removed)

        intersect = our_set.intersection(other_set)
        for res in intersect:
            # generate diff for each resource
            cmp = self._resources[res].compare(other._resources[res])
            if cmp:
                result.append(cmp)
            elif include_unmodified:
                result.append(self._resources[res].unmodified())

        return sorted(result, key=lambda r: r.resource_id)


def generate_diff(
    from_version_resources: Dict[ResourceIdStr, Resource],
    to_version_resources: Dict[ResourceIdStr, Resource],
    include_unmodified: bool = False,
) -> List[ResourceDiff]:
    """Generate a diff of two sets of resources, describing what has changed between them

    :param from_version_resources: The resources that are considered the starting point for comparison
    :param to_version_resources: The resources that are considered the target for comparison
    :param include_unmodified: If set to true,
    resources that haven't changed between the versions will be included in the results
    """
    from_state = Version(from_version_resources)
    to_state = Version(to_version_resources)

    return to_state.generate_diff(from_state, include_unmodified)



from setuptools import setup, find_namespace_packages
from os import path

requires = [
    "asyncpg~=0.25,<0.28",
    "click-plugins~=1.0",
    # click has been known to publish non-backwards compatible minors in the past (removed deprecated code in 8.1.0)
    "click>=8.0,<8.2",
    "colorlog~=6.4",
    "cookiecutter>=1,<3",
    "crontab>=0.23,<2.0",
    "cryptography>=36,<42",
    # docstring-parser has been known to publish non-backwards compatible minors in the past
    "docstring-parser>=0.10,<0.16",
    "email-validator>=1,<3",
    "execnet~=1.0",
    "importlib_metadata>=4,<7",
    "jinja2~=3.0",
    "more-itertools>=8,<10",
    "netifaces~=0.11",
    # leave upper bound floating for fast-moving and extremely stable packaging
    "packaging>=21.3",
    # pip>=21.3 required for editable pyproject.toml + setup.cfg based install support
    "pip>=21.3",
    "ply~=3.0",
    # lower bound because of pydantic/pydantic#5821
    "pydantic>=1.10.8,<2",
    "pyformance~=0.4",
    "PyJWT~=2.0",
    "pynacl~=1.5",
    "python-dateutil~=2.0",
    "pyyaml~=6.0",
    "texttable~=1.0",
    "tornado~=6.0",
    # lower bound because of ilevkivskyi/typing_inspect#100
    "typing_inspect~=0.9",
    "build~=0.7",
    "ruamel.yaml~=0.17",
    "toml~=0.10 ",
]


# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

version = "8.5.0"

setup(
    version=version,
    python_requires=">=3.9",  # also update classifiers
    # Meta data
    name="inmanta-core",
    description="Inmanta deployment tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Inmanta",
    author_email="code@inmanta.com",
    url="https://github.com/inmanta/inmanta-core",
    license="Apache Software License 2",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="orchestrator orchestration configurationmanagement",
    project_urls={
        "Bug Tracker": "https://github.com/inmanta/inmanta-core/issues",
        "Documentation": "https://docs.inmanta.com/community/latest/",
    },
    # Packaging
    package_dir={"": "src"},
    # All data files should be treated as namespace package according to
    # https://setuptools.pypa.io/en/latest/userguide/datafiles.html#subdirectory-for-data-files
    packages=find_namespace_packages(where="src"),
    # https://www.python.org/dev/peps/pep-0561/#packaging-type-information
    zip_safe=False,
    include_package_data=True,
    install_requires=requires,
    extras_require={
        "debug": ["rpdb"],
        # option to install a matched pair of inmanta-core and pytest-inmanta-extensions
        "pytest-inmanta-extensions": [f"pytest-inmanta-extensions~={version}.0.dev"],
        "datatrace": ["graphviz"],
    },
    entry_points={
        "console_scripts": [
            "inmanta-cli = inmanta.main:main",
            "inmanta = inmanta.app:app",
            "inmanta-initial-user-setup = inmanta.user_setup:main",
        ],
    },
)
